#include <node.h>
#include <Windows.h>

using namespace v8;

static DWORD s_lastErrorCode = S_OK;

bool hasError()
{
	return (s_lastErrorCode = GetLastError()) != S_OK;
}

void ThrowWin32Error(Isolate* isolate)
{
	wchar_t buf[1024] = {};
	FormatMessageW(FORMAT_MESSAGE_FROM_SYSTEM, NULL, s_lastErrorCode, LANG_USER_DEFAULT, buf, 1024, NULL);
	isolate->ThrowException(String::NewFromTwoByte(isolate, (const uint16_t*)buf));
}

template<size_t size>
static bool GetUnicodeArg0(const FunctionCallbackInfo<Value> &args, wchar_t (&buf)[size])
{
	Isolate* isolate = args.GetIsolate();
	String::Utf8Value utf8Str(isolate, args[0]);
	int length = MultiByteToWideChar(CP_UTF8, NULL, *utf8Str, utf8Str.length(), NULL, 0);
	if (length > size)
	{
		char buf[1024] = {};
		printf("Test");
		sprintf_s(buf, u8"ܳ%zuַ", size);
		isolate->ThrowException(String::NewFromUtf8(isolate, buf));
		return false;
	}

	MultiByteToWideChar(CP_UTF8, NULL, *utf8Str, utf8Str.length(), buf, LF_FACESIZE);
	return true;
}

struct EnumFontsStruct
{
	Isolate* isolate;
	Local<Array> fonts;
	int length;
};

struct FindFontsStruct
{
	wchar_t faceName[LF_FACESIZE];
	bool value;
};

static int CALLBACK FindFontsProc(
	const LOGFONTW    *lplf,
	const TEXTMETRICW *lptm,
	DWORD      dwType,
	LPARAM     lpData)
{
	FindFontsStruct* findFont = (FindFontsStruct*)lpData;
	if (wcscmp(findFont->faceName, lplf->lfFaceName) == 0)
	{
		findFont->value = true;
		return FALSE;
	}
	return TRUE;
}

static int CALLBACK EnumFontsProc(
	const LOGFONTW    *lplf,
	const TEXTMETRICW *lptm,
	DWORD      dwType,
	LPARAM     lpData
)
{
	EnumFontsStruct* fonts = (EnumFontsStruct*)lpData;
	fonts->fonts->Set(fonts->length++, String::NewFromTwoByte(fonts->isolate, (const uint16_t*)lplf->lfFaceName));
	return TRUE;
}

void listFonts(const FunctionCallbackInfo<Value> &args)
{
	Isolate *isolate = args.GetIsolate();
	EnumFontsStruct fonts;
	fonts.isolate = isolate;
	fonts.length = 0;
	fonts.fonts = Array::New(isolate);

	HDC hdc = GetDC(NULL);
	EnumFontsW(hdc, NULL, EnumFontsProc, (LPARAM)&fonts);
	ReleaseDC(NULL, hdc);
	fonts.fonts->Set(String::NewFromOneByte(isolate, (const uint8_t*)"length", NewStringType::kNormal).ToLocalChecked(),
		Int32::New(isolate, fonts.length));
	args.GetReturnValue().Set(fonts.fonts);
}

void existsFont(const FunctionCallbackInfo<Value> &args)
{
	Isolate *isolate = args.GetIsolate();
	if (args.Length() < 1 || !args[0]->IsString())
	{
		isolate->ThrowException(String::NewFromUtf8(isolate, u8"Ƿ."));
		return;
	}

	FindFontsStruct findFonts = {};
	if (!GetUnicodeArg0(args, findFonts.faceName))
	{
		args.GetReturnValue().Set(Boolean::New(isolate, false));
		return;
	}
	findFonts.value = false;
	HDC hdc = GetDC(NULL);
	EnumFontsW(hdc, NULL, FindFontsProc, (LPARAM)&findFonts);
	ReleaseDC(NULL, hdc);

	args.GetReturnValue().Set(Boolean::New(isolate, findFonts.value));
}

void addFont(const FunctionCallbackInfo<Value> &args)
{
	Isolate *isolate = args.GetIsolate();
	if (args.Length() < 1 || !args[0]->IsString())
	{
		isolate->ThrowException(String::NewFromUtf8(isolate, u8"Ƿ."));
		return;
	}
	wchar_t buf[MAX_PATH] = {};
	if (!GetUnicodeArg0(args, buf))
		return;

	bool result = AddFontResourceW(buf) != FALSE;
	//ʧܣwin32׳쳣
	if (!result && hasError())
		return ThrowWin32Error(isolate);
	args.GetReturnValue().Set(Boolean::New(isolate, result));
}

void removeFont(const FunctionCallbackInfo<Value> &args)
{
	Isolate *isolate = args.GetIsolate();
	if (args.Length() < 1 || !args[0]->IsString())
	{
		isolate->ThrowException(String::NewFromUtf8(isolate, u8"Ƿ."));
		return;
	}

	wchar_t buf[LF_FACESIZE] = {};
	if (!GetUnicodeArg0(args, buf))
		return;

	bool result = RemoveFontResourceW(buf) != FALSE;
	//ʧܣwin32׳쳣
	if (!result && hasError())
		return ThrowWin32Error(isolate);
	args.GetReturnValue().Set(Boolean::New(isolate, result));
}

void init(Local<Object> exports)
{
	NODE_SET_METHOD(exports, "listFonts", listFonts);
	NODE_SET_METHOD(exports, "addFont", addFont);
	NODE_SET_METHOD(exports, "removeFont", removeFont);
	NODE_SET_METHOD(exports, "existsFont", existsFont);
}

NODE_MODULE(NODE_GYP_MODULE_NAME, init)